#!/usr/bin/perl -w

##############################
#                            #
# SimpleWebServer/0.11       #
# -------------------------- #
# 0.11                       #
#   Fixes for binmode        #
#   Fixed MIME bug           #
#   Correct Content-length   #
#   Added several MIME Types #
# 0.1                        #
#   First version at all     #
#                            #
##############################

use strict;
use IO::Socket qw(:DEFAULT :crlf); # TCP/IP-Sockets
use HTTP::Date;                    # HTTP-Header-Datum
use POSIX qw(uname);               # Betriebssystem ermitteln

# --------------------------------

# Wichtige Konfigurationsvariablen; nicht nderbar
my $os = (uname)[0];               # Betriebssystem
my $sw = "SimpleWebServer";        # Servername
my $ver = "0.11";                  # Version
my $server = "$sw/$ver ($os)";
my %mime_types = (htm => 'text/html', html => 'text/html',
                  css => 'text/css', js => 'text/javascript',
                  txt => 'text/plain', gif => 'image/gif',
                  jpg => 'image/jpeg', png => 'image/png',
                  swf => 'application/x-shockwave-flash');

# Zeilenende-Kompatibilitt herstellen                  
my $crlf = CRLF;
$/ = $crlf;                        # Zeilenende auf CRLF
                  
# --------------------------------

# Konfiguration von der Kommandozeile lesen oder auf Standardwerte setzen

if ($ARGV[0] && $ARGV[0] eq "-h") {
   # Nur Hilfe abgefragt
   print << "ENDHELP";
$sw/$ver ($os)

Usage: $0 -h | -v | [PORT [SERVER_DIRECTORY [DIRECTORY_INDEX]]]

-h                Print this help message and exit
-v                Print version information and exit

PORT              Listening Port (Default: 8000)
SERVER_DIRECTORY  Public Web Server Directory (Default: ./htdocs)
DIRECTORY_INDEX   Name of Index File (Default: index.html)
ENDHELP
   exit (0);
} elsif ($ARGV[0] && $ARGV[0] eq "-v") {
   # Nur Version abgefragt
   print << "ENDVERSION";
$sw/$ver ($os)

Written by Sascha Kersken (sk\@lingoworld.de)
For the Galileo Press book 'Apache 2' (2004)
http://buecher.lingoworld.de/apache2

This program may be used, distributed and modified
under terms of the GNU General Public License
(See file COPYING in this directory)
ENDVERSION
   exit (0);
}

# Normalfall: Server soll ausgefhrt werden
   
# $port: Server-TCP-Port
# 1. Kommandozeilenargument
# Standard: 8000
my $port = $ARGV[0] || 8000;

# $serverdir: Server-Wurzelverzeichnis
# 2. Kommandozeilenargument
# Standard: Unterverzeichnis "htdocs" des aktuellen Verzeichnisses
my $serverdir = $ARGV[1] || "htdocs";

# $dirindex: Name der Indexdatei in jedem Verzeichnis
# 3. Kommandozeilenargument
# Standard: index.html
my $dirindex = $ARGV[2] || "index.html";

# --------------------------------

# Logdatei initialisieren
open LOG, ">>access.log";

# Lauschendes Socket erzeugen
my $listen = IO::Socket::INET->new
     (Proto     => 'tcp',
      LocalPort => $port,
      Listen    => SOMAXCONN);

print "$server laeuft auf Port $port ...\n\n";
print "Serververzeichnis: $serverdir\n";
print "Indexdokument:     $dirindex\n";

# Accept-Schleife
while (1) {
   # Weiter, falls keine Anfrage eintrifft
   next unless my $conn = $listen->accept();
   
   # Anfrage eingetroffen - Child-Prozess erzeugen
   my $child = fork;
   if ($child == 0) {
      # Im Child-Prozess Verbindung aufbauen und Anfrage beantworten
      handle_conn ($conn);
   }
}

sub handle_conn
{
   my $sock = shift;
   binmode $sock;
   my $peer_addr = $sock->peerhost;
   
   # Client-Anfrage lesen
   my $request = <$sock>;
   $request =~ s/\s*$//;
   my ($method, $uri, $http) = split (/\s+/, $request);
   
   # URI auf das Server-Verzeichnis umsetzen
   $uri =~ s|^/?|$serverdir/|;
   
   # Wenn URI ein Verzeichnis ist, Indexdatei verwenden
   if (-d $uri) {
      $uri =~ s|/?$|/$dirindex|;
   }
   
   # Die verschiedenen Datumsformate fr Logdatei und Serverantwort erzeugen
   my @months = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
   my $tm = time();
   my ($lsec, $lmin, $lhour, $lday, $lmon, $lyear) = localtime ($tm);
   my ($gsec, $gmin, $ghour, $gday, $gmon, $gyear) = gmtime ($tm);
   my $diff = $lhour - $ghour;
   my $logdate = "$gday/";
   $logdate .= $months[$gmon];
   $logdate .= "/";
   $logdate .= $gyear + 1900;
   $logdate .= ":";
   $logdate .= ($ghour < 10) ? "0$ghour:" : "$ghour:";
   $logdate .= ($gmin < 10) ? "0$gmin:" : "$gmin:";
   $logdate .= ($gsec < 10) ? "0$gsec" : "$gsec";
   $logdate .= ($diff < 0) ? " -" : " +";
   $diff = abs $diff;
   $logdate .= ($diff < 10) ? "0${diff}00" : "${diff}00";
   my $pagedate = time2str ($tm);
   
   # Weitere Elemente der Serverantwort
   my ($statuscode, $statusmsg, @headers, $body, $size);
   
   if ($method ne 'GET') {
      # Es werden NUR GET-Anfragen beantwortet
      $statuscode = 501;
      $statusmsg = "Not Implemented";
      $body = << "END_NI_BODY";     
<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//Final//EN\">
<html>
<head>
<title>501 Not Implemented</title>
</head>
<body>
The request method $method is not supported by this server.<br>
Server: $server<br>
$pagedate
</body>
</html>
END_NI_BODY
$size = length ($body);
      push (@headers, "Allow: GET");
      push (@headers, "Date: $pagedate");
      push (@headers, "Server: $server");
      push (@headers, "Connection: close");
      push (@headers, "Content-Type: text/html");
      push (@headers, "Content-Length: $size");
   } elsif (!(-e $uri)) {
      # Angeforderte Datei existiert nicht
      $statuscode = 404;
      $statusmsg = "Not Found";
      $body = << "END_NF_BODY";
<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//Final//EN\">
<html>
<head>
<title>404 Not Found</title>
</head>
<body>
The requested resource $uri could not be found on this server.<br>
Server: $server<br>
$pagedate
</body>
</html>
END_NF_BODY
$size = length ($body);
      push (@headers, "Date: $pagedate");
      push (@headers, "Server: $server");
      push (@headers, "Connection: close");
      push (@headers, "Content-Type: text/html");
      push (@headers, "Content-Length: $size");
   } else {
      # Datei existiert => lesen und zurckgeben
      # Wird der MIME-Type der angeforderten Datei untersttzt?
      # Standardtyp: text/plain
      my $mime_type = "text/plain";
      my $mime_ok = 0;
      foreach (keys %mime_types) {
         if ($uri =~ /\.$_$/) {
            $mime_ok = 1;
            $mime_type = $mime_types{$_};
         }
      }
      # Dateigre fr Content-length ermitteln
      $size = -s $uri;
      # Angeforderte Datei zeilenweise in @body einlesen
      open (IN, "<$uri") || warn ("Kann Datei $uri nicht oeffnen!");
      binmode IN;
      my $buffer;
      while (read (IN, $buffer, 16384)) {
         $body .= $buffer;
      }
      close (IN);
      $statuscode = 200;
      $statusmsg = "OK";
      push (@headers, "Date: $pagedate");
      push (@headers, "Server: $server");
      push (@headers, "Connection: close");
      push (@headers, "Content-Type: $mime_type");
      push (@headers, "Content-Length: $size");
   }
   
   # Ausgabe in die Logdatei
   print LOG "$peer_addr - - [$logdate] \"$request\" $statuscode $size\n";
   
   # Ausgabe der Daten an den Client
   print $sock "$http $statuscode $statusmsg$crlf";
   foreach my $header (@headers) {
      print $sock "$header$crlf";
   }
   print $sock "$crlf";
   print $sock "$body";
   close $sock;
   exit (0);
}